; Display board firmware, to allow the JW-002 display board to be
; used with the PJRC MP3 Player:   http://www.pjrc.com/tech/mp3/
; The JW-002 display board may be purchased from PJRC.COM, with
; this firmware installed and an adaptor board with 12 pushbuttons
; and two cables to allow it to connect directly to the PJRC MP3
; player circuit board's 4-pin connector.

; The JW-002 display board may also be purchased directly from
; Earth Computer Tech, at:  http://www.flat-panel.com/jw002.htm
; Earth will offer quantity discounts, so for applications other
; than use with the PJRC MP3 player, Earth may be a better choice
; for purchasing the display.  They provide the JW-002 without
; the extra parts needed for use with the PJRC MP3 player board.

; This JW-002 firmware source code has been placed in the public domain.
; This code is provided in the hope that it will be useful, but
; it is provided without any warranty; without even the implied
; warranty of merchantability or fitness for a particular purpose. 


;TODO: add pushbutton reading code
;TODO: implement control characters


	;configurable (but hard-coded) settings

	.equ    location, 0x0000	;stand-alone build
	;.equ    location, 0x8000	;to run under PAULMON2 (see below)

	;.equ	baud_const, 255		;57600 baud
	.equ	baud_const, 253		;19200 baud
	;.equ	baud_const, 250		;9600 baud


; ******************************************************************



	;internal memory allocation
	.equ	tx_buf, 0x10		;12 byte transmit buffer
	.equ	tx_buf_size, 0x0C
	.equ	rx_buf_head, 0x1C	;1 byte
	.equ	rx_buf_tail, 0x1D	;1 byte
	.equ	tx_buf_head, 0x1E	;1 byte
	.equ	tx_buf_tail, 0x1F	;1 byte
	.equ	mode_bits, 0x20		;1 byte
	.equ	pb_bits, 0x21		;2 bytes
	.equ	x_pos, 0x23		;1 byte, current x char position
	.equ	y_pos, 0x24		;1 byte, current y char position
	.equ	char_buf, 0x25		;5 bytes, buffer of current char
	.equ	font, 0x2A		;1 byte, current font to use
	.equ	cmd_len, 0x2B		;1 byte, # of chars revc
	.equ	scroll_offset, 0x2C	;1 byte, current y scroll offset
	.equ	pb_scan_state, 0x2D	;1 byte
	.equ	cmd_buf, 0x30		;16 byte command buffer
	.equ	rx_buf, 0x40		;40 byte receive buffer
	.equ	rx_buf_size, 0x28
	.equ	stack, 0x68		;24 bytes left for the stack

	;bit mapped memory allocation
	.equ	wrap_mode, 0
	.equ	scroll_mode, 1
	.equ	ignore_input_mode, 2
	.equ	cin_serviced, 3
	.equ	pb_bit_0, 8
	.equ	pb_bit_1, 9
	.equ	pb_bit_2, 10
	.equ	pb_bit_3, 11
	.equ	pb_bit_4, 12
	.equ	pb_bit_5, 13
	.equ	pb_bit_6, 14
	.equ	pb_bit_7, 15
	.equ	pb_bit_8, 16
	.equ	pb_bit_9, 17
	.equ	pb_bit_A, 18
	.equ	pb_bit_B, 19

	;memory mapped interface to the LCD display controller chips
	.equ	left_cmd, 0xFF00
	.equ	left_data, 0xFF02
	.equ	right_cmd, 0xFF04
	.equ	right_data, 0xFF06

	;connections to the 8051 port pins
        .equ    pb_xmit_1, 0xB2         ;p3.2
        .equ    pb_xmit_2, 0xB3         ;p3.3
        .equ    pb_xmit_3, 0x97         ;p1.7
        .equ    pb_recv_1, 0x93         ;p1.3
        .equ    pb_recv_2, 0x92         ;p1.2
        .equ    pb_recv_3, 0x94         ;p1.4
        .equ    pb_recv_4, 0x90         ;p1.0
	.equ	led_pin, 0x91		;clear to turn on LED
	.equ	watchdog_pin, 0xB5	;toggle to clear watchdog timer






;***********************************************************************
;**                                                                   **
;**                          Main Program                             **
;**                                                                   **
;***********************************************************************

	.org	location
	ljmp	begin

	.org	location+11
	ljmp	timer0_isr

	.org	location+35
	ljmp	uart_intr

.org    location+256
.db     0xA5,0xE5,0xE0,0xA5     ;signiture bytes
.db     35,255,0,0              ;id
.db     0,0,0,0                 ;reserved
.db     0,0,0,0                 ;reserved
.db     0,0,0,0                 ;reserved
.db     0,0,0,0                 ;reserved
.db     0,0,0,0                 ;user defined
.db     255,255,255,255         ;length and checksum (255=unused)
.db     "JW-002 LCD Firmware",0
.org    location+256+64         ;executable code begins here


; To allow for program development with PAULMON2, the JW-002 display
; board needs several modifications.
;
; 1: Program PAULMON2 (pm21_8.obj) into the flash chip.
;
; 2: Disable the watchdog timer: short C1.
;
; 3: Disconnect the flash chip's CE pin from ground and connect it to
;    A15.  The wide trace that runs under the upper left corner of the 
;    PLCC socket is ground and only connects to CE, so this is the
;    trace to cut.  A15 may be found on a via 0.25 inches away from
;    the corner of this trace near the PLCC socket.
;
; 4: Connect PSEN and RD to an AND gate.  The two unused NAND gates
;    in IC2 may be used, if the pins are lifted away from the board.
;
; 5: Disconnect IC7's CE (pin 22) and connect it to the AND gate output.
;
; With these changes, this code may be built at 8000 and downloaded to
; the board.  Memory from 0000 to 7FFF will be both code memory (the
; flash chip with PAULMON2) and data memory (32k SRAM, IC8), and the
; firmware will use the SRAM starting at 0000 for fonts.  The code
; allows 32 fonts, but care should be taken not to use more than the
; first 25 fonts, as the 26th font will begin to overwrite the program
; at 8000.


begin:
	mov	sp, #stack
	lcall	copy_fonts_to_sram
	acall	init_display

	acall	move_cursor_home
	mov	font, #0
	mov	cmd_len, #0
	setb	wrap_mode
	setb	scroll_mode
	clr	ignore_input_mode

	lcall	init_uart_intr
	lcall	timer0_setup

	mov	a, #location >> 8
	jz	main_loop
	mov	dptr, #mesg_abort
	lcall	pstr


main_loop:
	lcall	cin
	;lcall	cout			;uncomment for echo of input characters
	mov	r2, a
main1:	mov	a, cmd_len		;check if continuing multi-byte cmd
	jz	main2
	;if cmd_len > 0, this byte is to be treated at part of
	;a multi-byte command, and not regular input
	mov	a, #cmd_buf
	add	a, cmd_len
	mov	r0, a
	mov	a, r2
	mov	@r0, a
	inc	cmd_len
	acall	exec_cmd
	ajmp	main_loop
main2:	mov	a, r2			;check for 0x5C command
	cjne	a, #0x5C, main3
	;the '\' character, 0x5C, begin a command
	mov	r0, #cmd_buf
	mov	a, r2
	mov	@r0, a
	mov	cmd_len, #1
	;all '\' commands are at least two characters, so don't call exec_cmd
	ajmp	main_loop
main3:	mov	a, r2			;check for control character
	add	a, #224
	jc	main4
	mov	r0, #cmd_buf
	mov	a, r2
	mov	@r0, a
	mov	cmd_len, #1
	acall	exec_cmd
	ajmp	main_loop
main4:	jb	ignore_input_mode, main_loop	;check ignore input mode

main_draw:
	mov	a, r2
	lcall	lookup_char
	acall	draw_char
	ajmp	main_loop




;***********************************************************************
;**                                                                   **
;**                     Handle Control Characters                     **
;**                                                                   **
;***********************************************************************



exec_cmd:
	mov	r0, #cmd_buf
	mov	a, @r0
	cjne	a, #0x5C, exec_ctrl
	ajmp	exec_5c_cmd

exec_ctrl:

	;ESC Sequence
escape:	cjne	a, #27, ctrl_a
	ajmp	exec_esc

	;CTRL-A: Cursor Home
ctrl_a: cjne	a, #1, ctrl_h
	mov	cmd_len, #0
	ret

	;CTRL-H: Backspace
ctrl_h:	cjne	a, #8, ctrl_i
	mov	cmd_len, #0
	ret

	;CTRL-I: Tab
ctrl_i:	cjne	a, #9, ctrl_j
	mov	cmd_len, #0
	ret

	;CTRL-J: Line Feed
ctrl_j:	cjne	a, #10, ctrl_l
	mov	cmd_len, #0
	ret

;TODO: add ctrl-k as up one line

	;CTRL-L: Form Feed (clear display)
ctrl_l:	cjne	a, #12, ctrl_m
	mov	cmd_len, #0
	ret

	;CTRL-M: Carriage Return
ctrl_m:	cjne	a, #13, ctrl_u
	mov	cmd_len, #0
	ret

	;CTRL-U: Forward
ctrl_u:	cjne	a, #21, crtl_unknown
	mov	cmd_len, #0
	ret

crtl_unknown:
	;just ignore unknown control characters
	mov	cmd_len, #0
	ret



;***********************************************************************
;**                                                                   **
;**                       Handle Escape Sequences                     **
;**                                                                   **
;***********************************************************************


	;0x1B 0x5B 0x41, move up one line
	;0x1B 0x5B 0x42, move down one line
	;0x1B 0x5B 0x43, move right one line
	;0x1B 0x5B 0x44, move left one line
	;0x1B 0x5B 0x48, cursor home
	;0x1B 0x5B 0x30 0x3B 0x37 0x6B, begin inverse text
	;0x1B 0x5B 0x30 0x6B, begin normal text
	;0x1B 0x5B 0x32 0x4A, clear screen

exec_esc:
	mov	a, cmd_len
	add	a, #254
	jc	exec_esc_1
	ret				;if len < 2, wait for more

exec_esc_1:
	mov	a, cmd_len
	add	a, #240
	jnc	exec_esc_2		;check if len >= 16
	mov	cmd_len, #0		;if len > 16, ignore input
	ret

exec_esc_2:
	mov	r0, #cmd_buf+1
	mov	a, @r0
	cjne	a, #0x5B, exec_esc_esc
	ajmp	exec_esc_5B

exec_esc_esc:
	cjne	a, #0x1B, exec_esc_unknown
	mov	a, #location >> 8
	jnz	exec_esc_abort
	dec	cmd_len			;ignore the 2nd ESC
	ret
exec_esc_abort:
	ljmp	0			;return to PAULMON2 on two ESCs


exec_esc_unknown:
	;just ignore unknown escape sequences
	mov	cmd_len, #0
	ret

exec_esc_5B:
	;TODO: this parsing really ought to allow numeric and ';'
	;characters and conclude that we have a complete escape
	;sequence when we get a alpha character A-Z or a-z.
	mov	a, cmd_len
	add	a, #253
	jc	exec_esc_5B_2
	ret				;if len < 3, wait for more

exec_esc_5B_2:
	mov	r0, #cmd_buf+2
	mov	a, @r0

exec_esc_5B_41:
	cjne	a, #0x41, exec_esc_5B_42
	mov	a, #1
	acall	move_cursor_up
	mov	cmd_len, #0
	ret

exec_esc_5B_42:
	cjne	a, #0x42, exec_esc_5B_43
	mov	a, #1
	acall	move_cursor_down
	mov	cmd_len, #0
	ret

exec_esc_5B_43:
	cjne	a, #0x43, exec_esc_5B_44
	mov	a, #1
	acall	move_cursor_right
	mov	cmd_len, #0
	ret

exec_esc_5B_44:
	cjne	a, #0x44, exec_esc_5B_unknown
	mov	a, #1
	acall	move_cursor_left
	mov	cmd_len, #0
	ret

exec_esc_5B_unknown:
	;just ignore unknown escape sequences
	mov	cmd_len, #0
	ret





;***********************************************************************
;**                                                                   **
;**                   Handle Custom Display Commands                  **
;**                                                                   **
;***********************************************************************


exec_5c_cmd:
	mov	a, cmd_len
	add	a, #254
	jc	exec_5c_cmd_1
	ret				;if len < 2, wait for more

exec_5c_cmd_1:
	mov	a, cmd_len
	add	a, #240
	jnc	exec_5c_cmd_2		;check if len >= 16
	mov	cmd_len, #0		;if len > 16, ignore input
	ret


	;cmd 0x5C <0x20-0x3F>, display bytes 0 to 31
exec_5c_cmd_2:
	mov	r0, #cmd_buf+1
	mov	a, @r0
	anl	a, #11100000b
	cjne	a, #00100000b, exec_5c_cmd_3
	;if we get here, the cmd is to display bytes 0 to 31
	mov	a, @r0
	add	a, #224
exec_display_char:
	lcall	lookup_char
	acall	draw_char
	mov	cmd_len, #0
	ret


	;cmd 0x5C <0x60-0x7F>, display bytes 128 to 159
exec_5c_cmd_3:
	cjne	a, #01100000b, exec_5c_5c
	;if we get here, the cmd is to display bytes 128 to 159
	mov	a, @r0
	add	a, #32
	sjmp	exec_display_char


	;cmd 0x5C 0x5C, display '\'
exec_5c_5c:
	mov	a, @r0
	cjne	a, #0x5C, exec_5c_5f
	sjmp	exec_display_char


	;cmd 0x5C 0x5C, display byte 127
exec_5c_5f:
	cjne	a, #0x5F, exec_5c_5d
	mov	a, #127
	sjmp	exec_display_char


	;cmd 0x5C 0x5D, begin ignoring input bytes (but not commands)
exec_5c_5d:
	cjne	a, #0x5D, exec_5c_5b
	setb	ignore_input_mode
	mov	cmd_len, #0
	ret


	;cmd 0x5C 0x5B, don't ignore input bytes
exec_5c_5b:
	cjne	a, #0x5B, exec_5c_40
	clr	ignore_input_mode
	mov	cmd_len, #0
	ret


	;cmd 0x5C 0x40 <font+32> <opcode>, clear display or show font summary
exec_5c_40:
	cjne	a, #0x40, exec_5c_41
	mov	a, cmd_len
	cjne	a, #4, exec_5c_40_end
	mov	r0, #cmd_buf+2
	mov	a, @r0
	acall	byte2font
	jc	exec_5c_40_done
	mov	font, a
	mov	r0, #cmd_buf+3
	mov	a, @r0
exec_5c_40_30:
	cjne	a, #0x30, exec_5c_40_31
	acall	init_display
	sjmp	exec_5c_40_done
exec_5c_40_31:
	cjne	a, #0x31, exec_5c_40_32
	mov	r6, #0
	mov	r0, #cmd_buf+4
	mov	@r0, #msg_f0 & 255
	inc	r0
	mov	@r0, #msg_f0 >> 8
	sjmp	exec_5c_40_font_preview
exec_5c_40_32:
	cjne	a, #0x32, exec_5c_40_done
	mov	r6, #128
	mov	r0, #cmd_buf+4
	mov	@r0, #msg_f128 & 255
	inc	r0
	mov	@r0, #msg_f128 >> 8
exec_5c_40_font_preview:		;draw font preview page
	lcall	test_font
	mov	a, font
	add	a, #'0'
	mov	r6, a
	push	font
	mov	font, #0
	acall	move_cursor_home
	mov	dptr, #msg_font
	acall	lcd_pstr
	mov	a, r6
	lcall	lookup_char
	acall	draw_char
	mov	x_pos, #0
	mov	y_pos, #2
	mov	r0, #cmd_buf+4
	mov	dpl, @r0
	inc	r0
	mov	dph, @r0
	acall	lcd_pstr
	mov	x_pos, #0
	mov	y_pos, #3
	acall	lcd_pstr
	pop	font
exec_5c_40_done:
	acall	move_cursor_home
	mov	cmd_len, #0
exec_5c_40_end:
	ret


	;cmd 0x5C 0x41 <mode>, set modes (wrap and scroll)
exec_5c_41:
	cjne	a, #0x41, exec_5c_42
	mov	a, cmd_len
	cjne	a, #3, exec_5c_41_end
	mov	r0, #cmd_buf+2
	mov	a, @r0
	anl	a, #11110000b
	cjne	a, #0x30, exec_5c_41_done
	mov	a, @r0
	rrc	a
	mov	wrap_mode, c
	rrc	a
	mov	scroll_mode, c
exec_5c_41_done:
	mov	cmd_len, #0
exec_5c_41_end:
	ret


	;cmd 0x5C 0x42 <column+32> <row+32>, set cursor position
exec_5c_42:
	cjne	a, #0x42, exec_5c_43
	mov	a, cmd_len
	cjne	a, #4, exec_5c_42_end
	mov	r0, #cmd_buf+2
	mov	a, @r0
	add	a, #224
	mov	x_pos, a
	mov	r0, #cmd_buf+3
	mov	a, @r0
	add	a, #224
	mov	y_pos, a
exec_5c_42_done:
	mov	cmd_len, #0
exec_5c_42_end:
	ret


	;cmd 0x5C 0x43 <font+32>, set font
exec_5c_43:
	cjne	a, #0x43, exec_5c_44
	mov	a, cmd_len
	cjne	a, #3, exec_5c_43_end
	mov	r0, #cmd_buf+2
	mov	a, @r0
	acall	byte2font
	jc	exec_5c_43_done
	mov	font, a
exec_5c_43_done:
	mov	cmd_len, #0
exec_5c_43_end:
	ret




	;define a custom character
	;cmd 0x5C 0x44 <char (2 bytes hex)> <font+32> <bitmap (8 bytes)>
	;
	;  example: replace "R" (R is ascii code 0x52) in font 0
	;  0x5C 0x44 0x35 0x32 0x20 0x3E 0x39 0x39 0x39 0x3E 0x3A 0x39 0x39
	;  as an ascii string: "\D52 >999>:99"
	;  
	; ****   00111110  0x3E
	; **  *  00111001  0x39
	; **  *  00111001  0x39
	; **  *  00111001  0x39
	; ****   00111110  0x3E
	; ** *   00111010  0x3A
	; **  *  00111001  0x39
	; **  *  00111001  0x39
exec_5c_44:
	cjne	a, #0x44, exec_5c_45
	mov	a, cmd_len
	cjne	a, #13, exec_5c_44_end
	mov	r0, #cmd_buf+2
	mov	a, @r0
	acall	asc2hex
	jc	exec_5c_44_done
	swap	a
	mov	r1, a
	mov	r0, #cmd_buf+3
	mov	a, @r0
	acall	asc2hex
	jc	exec_5c_44_done
	orl	a, r1
	mov	r1, a			;r1 is the character to write
	mov	r0, #cmd_buf+4
	mov	a, @r0
	acall	byte2font
	jc	exec_5c_44_done
	mov	r2, a			;r2 is the font number of change
	mov	r0, #cmd_buf+5
exec_5c_44_loop1:
	mov	a, @r0			;make sure all bitmap bytes are legal
	anl	a, #11100000b
	cjne	a, #00100000b, exec_5c_44_done
	inc	r0
	cjne	r0, #cmd_buf+13, exec_5c_44_loop1
	mov	a, r2
	mov	b, #5
	mul	ab			;compute font's memory position
	mov	r7, a
	mov	a, r1
	mov	b, #5
	mul	ab			;compute offset within font
	mov	dpl, a
	mov	a, b
	add	a, r7
	mov	dph, a			;dptr points to character to update
	mov	r3, #0			;r3 counts font x position (0 to 4)
exec_5c_44_loop2:
	mov	a, r3
	acall	exec_5c_44_bitmask
	mov	r4, a			;r4 holds bitmask for this x position
	mov	r0, #cmd_buf+5
	mov	r5, #0			;r5 will become 8 bit vertical pattern
exec_5c_44_loop3:
	mov	a, @r0
	anl	a, r4
	setb	c
	jnz	exec_5c_44_loop3b
	clr	c
exec_5c_44_loop3b:
	mov	a, r5
	rrc	a
	mov	r5, a
	inc	r0
	cjne	r0, #cmd_buf+13, exec_5c_44_loop3
	mov	a, r5
	movx	@dptr, a		;write to the in-memory font
	inc	dptr
	inc	r3
	cjne	r3, #5, exec_5c_44_loop2
exec_5c_44_done:
	mov	cmd_len, #0
exec_5c_44_end:
	ret
exec_5c_44_bitmask:
	inc	a
	movc	a, @a+pc
	ret
	.db	16, 8, 4, 2, 1


exec_5c_45:
	cjne	a, #0x45, exec_5c_46
	mov	a, #0xFF
	lcall	cout
	mov	a, #0xC2
	lcall	cout
	mov	cmd_len, #0
	ret


exec_5c_46:
	cjne	a, #0x46, exec_5c_47
	mov	a, cmd_len
	cjne	a, #3, exec_5c_46_end
	mov	r0, #cmd_buf+2
	mov	a, @r0
	add	a, #224
	jnc	exec_5c_46_done
exec_5c_46_left:
	add	a, #224
	jc	exec_5c_46_right
	add	a, #32
	acall	move_cursor_left
	sjmp	exec_5c_46_done
exec_5c_46_right:
	add	a, #224
	jc	exec_5c_46_down
	add	a, #32
	acall	move_cursor_right
	sjmp	exec_5c_46_done
exec_5c_46_down:
	add	a, #240
	jc	exec_5c_46_up
	add	a, #16
	acall	move_cursor_down
	sjmp	exec_5c_46_done
exec_5c_46_up:
	add	a, #240
	jc	exec_5c_46_up
	add	a, #16
	acall	move_cursor_up
	sjmp	exec_5c_46_done
exec_5c_46_done:
	mov	cmd_len, #0
exec_5c_46_end:
	ret




exec_5c_47:

exec_5c_unknown:
	lcall	phex
	mov	a, #'*'
	lcall	cout
	mov	cmd_len, #0
	ret





;***********************************************************************
;**                                                                   **
;**                           Misc Functions                          **
;**                                                                   **
;***********************************************************************




	;convert recv'd byte in Acc to a font #  (by subtracting 32)
	;C=0 if legal, C=1 if illegal
byte2font:
	add	a, #224
	jnc	b2f_nope
	add	a, #224
	jc	b2f_nope
	add	a, #32
	clr	c
	ret
b2f_nope:
	clr	a
	setb	c
	ret



        ;carry set if invalid input
asc2hex:
        add     a, #208
        jnc     hex_not
        add     a, #246
        jc      hex_maybe
        add     a, #10
        clr     c
        ret
hex_maybe:
        add     a, #249
        jnc     hex_not
        add     a, #250
        jc      hex_not
        add     a, #16
        clr     c
        ret
hex_not:setb    c
        ret


lcd_pstr:
	clr	a
	movc	a, @a+dptr
	inc	dptr
	jz	lcd_pstr_end
	push	dph
	push	dpl
	lcall	lookup_char
	acall	draw_char
	pop	dpl
	pop	dph
	sjmp	lcd_pstr
lcd_pstr_end:
	ret




;***********************************************************************
;**                                                                   **
;**                         Cursor Movement                           **
;**                                                                   **
;***********************************************************************


move_cursor_home:
	mov	x_pos, #0
	mov	y_pos, #0
	ret


	;move the cursor right by Acc positions
move_cursor_right:
	mov	r7, a			;keep num positions in r7
	add	a, x_pos
	add	a, #232
	jc	mov_rt_2
	add	a, #24
	mov	x_pos, a		;easy case, no wrap around
	ret
mov_rt_2:  ;the x movement puts the cursor off screen
	jb	wrap_mode, mov_rt_3
	mov	x_pos, #23		;right edge when wrap mode off
	ret
mov_rt_3:
	mov	a, #1
	acall	move_cursor_down
	mov	a, #24
	clr	c
	subb	a, x_pos
	mov	r0, a
	mov	a, r7
	clr	c
	subb	a, r0
	mov	x_pos, #0
	sjmp	move_cursor_right


	;move the cursor left by Acc positions
move_cursor_left:
	mov	r7, a			;keep num positions in r7
	mov	a, x_pos
	clr	c
	subb	a, r7
	jc	mov_lf_2
	mov	x_pos, a		;easy case, no wrap around
	ret
mov_lf_2:  ;the x movement puts the cursor off screen
	jb	wrap_mode, mov_lf_3
	mov	x_pos, #0		;left edge when wrap mode off
	ret
mov_lf_3:
	mov	a, #1
	acall	move_cursor_up
	mov	a, r7
	setb	c
	subb	a, x_pos
	mov	x_pos, #23
	sjmp	move_cursor_left


	;move the cursor down by Acc positions
move_cursor_down:
	mov	r6, a			;keep num positions in r6
	add	a, y_pos
	add	a, #248
	jc	mov_dn_2
	add	a, #8
	mov	y_pos, a		;easy case, didn't hit the bottom
	ret
mov_dn_2:
	jb	scroll_mode, mov_dn_sch
	jb	wrap_mode, mov_dn_wrap
	mov	y_pos, #7		;bottom edge if no wrap or scroll
	ret
mov_dn_sch:
	mov	a, #7
	clr	c
	subb	a, y_pos
	mov	r0, a
	mov	a, r6
	clr	c
	subb	a, r0
	mov	r6, a
	mov	y_pos, #7
	acall	scroll_normal
	dec	r6
	mov	a, r6
	jnz	move_cursor_down
	ret
mov_dn_wrap:
	mov	a, #8
	clr	c
	subb	a, y_pos
	mov	r0, a
	mov	a, r6
	clr	c
	subb	a, r0
	mov	y_pos, #0
	sjmp	move_cursor_down


	;move the cursor up by Acc positions
move_cursor_up:
	mov	r6, a			;keep num positions in r6
	mov	a, y_pos
	clr	c
	subb	a, r6
	jc	mov_up_2
	mov	y_pos, a		;easy case, didn't hit the top
	ret
mov_up_2:
	jb	scroll_mode, mov_up_sch
	jb	wrap_mode, mov_up_wrap
	mov	y_pos, #0		;top edge if no wrap or scroll
	ret
mov_up_sch:
	mov	a, r6
	clr	c
	subb	a, y_pos
	mov	r6, a
	mov	y_pos, #0
	acall	scroll_reverse
	dec	r6
	mov	a, r6
	jnz	move_cursor_up
	ret
mov_up_wrap:
	mov	a, r6
	setb	c
	subb	a, y_pos
	mov	y_pos, #7
	sjmp	move_cursor_up



;***********************************************************************
;**                                                                   **
;**             LCD Display Controller (KS0108B) Routines             **
;**                                                                   **
;***********************************************************************



draw_char:
	mov	a, x_pos
	add	a, #244
	jnc	draw_char_left
	add	a, #244
	jnc	draw_char_right
	ret


draw_char_left:
        acall	left_busy
	mov	dptr, #left_cmd
	mov	a, x_pos
	mov	b, #5
	mul	ab
	anl	a, #00111111b
	orl	a, #0x40
	movx	@dptr, a

        acall	left_busy
	mov	dptr, #left_cmd
	mov	a, scroll_offset
	swap	a
	rl	a
	add	a, y_pos
	anl	a, #00000111b
	orl	a, #0xB8
	movx	@dptr, a

        acall	left_busy
	mov	dptr, #left_data
	mov	a, char_buf+0
	movx	@dptr, a
	
        acall	left_busy
	mov	dptr, #left_data
	mov	a, char_buf+1
	movx	@dptr, a

        acall	left_busy
	mov	dptr, #left_data
	mov	a, char_buf+2
	movx	@dptr, a

        acall	left_busy
	mov	dptr, #left_data
	mov	a, char_buf+3
	movx	@dptr, a

        acall	left_busy
	mov	dptr, #left_data
	mov	a, char_buf+4
	movx	@dptr, a

	mov	a, #1
	acall	move_cursor_right
	ret


draw_char_right:
        acall   right_busy
	mov	dptr, #right_cmd
	mov	a, x_pos
	add	a, #244
	mov	b, #5
	mul	ab
	anl	a, #00111111b
	orl	a, #0x40
	movx	@dptr, a

        acall   right_busy
	mov	dptr, #right_cmd
	mov	a, scroll_offset
	swap	a
	rl	a
	add	a, y_pos
	anl	a, #00000111b
	orl	a, #0xB8
	movx	@dptr, a

        acall   right_busy
	mov	dptr, #right_data
	mov	a, char_buf+0
	movx	@dptr, a
	
        acall   right_busy
	mov	dptr, #right_data
	mov	a, char_buf+1
	movx	@dptr, a

        acall   right_busy
	mov	dptr, #right_data
	mov	a, char_buf+2
	movx	@dptr, a

        acall   right_busy
	mov	dptr, #right_data
	mov	a, char_buf+3
	movx	@dptr, a

        acall   right_busy
	mov	dptr, #right_data
	mov	a, char_buf+4
	movx	@dptr, a

	mov	a, #1
	acall	move_cursor_right
	ret





left_busy:
	mov	dptr, #left_cmd
lbusy2:	movx	a, @dptr
	jb	acc.7, lbusy2
	ret


right_busy:
	mov	dptr, #right_cmd
rbusy2:	movx	a, @dptr
	jb	acc.7, rbusy2
	ret



scroll_normal:
	mov	a, scroll_offset
	add	a, #8
	anl	a, #0x3F
	mov	scroll_offset, a
	acall	update_scroll_offset
	mov	r0, #7			;r0 is line # to clear
	sjmp	scroll_clear_line

scroll_reverse:
	mov	a, scroll_offset
	add	a, #248
	anl	a, #0x3F
	mov	scroll_offset, a
	acall	update_scroll_offset
	mov	r0, #0			;r0 is line # to clear

scroll_clear_line:
        acall   left_busy
	mov	dptr, #left_cmd
	mov	a, #0x40
	movx	@dptr, a
        acall   right_busy
	mov	dptr, #right_cmd
	mov	a, #0x40
	movx	@dptr, a
        acall   left_busy
	mov	dptr, #left_cmd
	mov	a, scroll_offset
	swap	a
	rl	a
	add	a, r0
	anl	a, #00000111b
	orl	a, #0xB8
	movx	@dptr, a
        acall   right_busy
	mov	dptr, #right_cmd
	mov	a, scroll_offset
	swap	a
	rl	a
	add	a, r0
	anl	a, #00000111b
	orl	a, #0xB8
	movx	@dptr, a
	mov	r3, #64			;r3 counts horizontal position
scroll_loop:
	acall   left_busy
	mov	dptr, #left_data
	clr	a
	movx	@dptr, a		;clear left side
	acall	right_busy
	mov	dptr, #right_data
	clr	a
	movx	@dptr, a		;clear right side
	djnz	r3, scroll_loop
	ret


update_scroll_offset:
	acall	left_busy
	mov	dptr, #left_cmd
	mov	a, scroll_offset
	anl	a, #0x3F
	orl	a, #0xC0
	movx	@dptr, a		;set top of display at y=0
	acall	right_busy
	mov	dptr, #right_cmd
	mov	a, scroll_offset
	anl	a, #0x3F
	orl	a, #0xC0
	movx	@dptr, a		;set top of display at y=0
	ret



init_display:
	mov	scroll_offset, #0
	acall	update_scroll_offset
	mov	r2, #0			;r2 counts line number
init2:	acall	left_busy
	mov	dptr, #left_cmd
	mov	a, #0x40
	movx	@dptr, a
	acall	right_busy
	mov	dptr, #right_cmd
	mov	a, #0x40
	movx	@dptr, a
	acall	left_busy
	mov	dptr, #left_cmd
	mov	a, r2
	orl	a, #0xB8
	movx	@dptr, a
	acall	right_busy
	mov	dptr, #right_cmd
	mov	a, r2
	orl	a, #0xB8
	movx	@dptr, a
	mov	r3, #64			;r3 counts horizontal position
init3:	acall   left_busy
	mov	dptr, #left_data
	clr	a
	movx	@dptr, a		;clear left side
	acall	right_busy
	mov	dptr, #right_data
	clr	a
	movx	@dptr, a		;clear right side
	djnz	r3, init3
	inc	r2
	cjne	r2, #8, init2
	acall	left_busy
	mov	dptr, #left_cmd
	mov	a, #0x3F
	movx	@dptr, a		;turn on display
	acall	right_busy
	mov	dptr, #right_cmd
	mov	a, #0x3F
	movx	@dptr, a		;turn on display
        ret



;***********************************************************************
;**                                                                   **
;**                           Pushbuttons                             **
;**                                                                   **
;***********************************************************************


timer0_setup:
	clr	ea
	setb	pb_recv_1
	setb	pb_recv_2
	setb	pb_recv_3
	setb	pb_recv_4
	clr	pb_xmit_1
	setb	pb_xmit_2
	setb	pb_xmit_3
	mov	pb_scan_state, #0
	orl	tmod, #00000001b
	anl	tmod, #11110001b
	clr	tr0
	clr	tf0
	clr	pt0
	setb	et0
	setb	ea
	setb	tr0
	ret


timer0_isr:
	push	psw
	push	acc
	clr	tr0
	mov	th0, #0xFA		;run this code at approx 600 Hz
	mov	tl0, #0x12
	setb	tr0
	jnb	cin_serviced, pb0
	;clear the watchdog timer, but only if the main program has been
	;running "cin" (accepting user input) and of course this timer
	;interrupt routine is still working.  If we were really paranoid,
	;there be some sort of check of see if the UART (and timer1) still
	;look like they're configured correctly
	cpl	watchdog_pin
	clr	cin_serviced
pb0:
	mov	a, pb_scan_state
pb1:	cjne	a, #0, pb2
	mov	c, pb_recv_1
	mov	pb_bit_0, c
	mov	c, pb_recv_2
	mov	pb_bit_1, c
	mov	c, pb_recv_3
	mov	pb_bit_2, c
	mov	c, pb_recv_4
	mov	pb_bit_3, c
	mov	pb_scan_state, #1
	setb	pb_xmit_1
	clr	pb_xmit_2
	setb	pb_xmit_3
	pop	acc
	pop	psw
	reti
pb2:	cjne	a, #1, pb3
	mov	c, pb_recv_1
	mov	pb_bit_4, c
	mov	c, pb_recv_2
	mov	pb_bit_5, c
	mov	c, pb_recv_3
	mov	pb_bit_6, c
	mov	c, pb_recv_4
	mov	pb_bit_7, c
	mov	pb_scan_state, #2
	setb	pb_xmit_1
	setb	pb_xmit_2
	clr	pb_xmit_3
	pop	acc
	pop	psw
	reti
pb3:	cjne	a, #2, pb4
	mov	c, pb_recv_1
	mov	pb_bit_8, c
	mov	c, pb_recv_2
	mov	pb_bit_9, c
	mov	c, pb_recv_3
	mov	pb_bit_A, c
	mov	c, pb_recv_4
	mov	pb_bit_B, c
	mov	pb_scan_state, #0
	clr	pb_xmit_1
	setb	pb_xmit_2
	setb	pb_xmit_3
	sjmp	pb_analysis
pb4:	mov	pb_scan_state, #0
	clr	pb_xmit_1
	setb	pb_xmit_2
	setb	pb_xmit_3
	pop	acc
	pop	psw
	reti

pb_analysis:

	;do something with the pushbutton bits here

	;clr	a
	;mov	c, pb_bit_0
	;addc	a, #'0'
	;lcall	cout
	;clr	a
	;mov	c, pb_bit_1
	;addc	a, #'0'
	;lcall	cout
	;clr	a
	;mov	c, pb_bit_2
	;addc	a, #'0'
	;lcall	cout
	;clr	a
	;mov	c, pb_bit_3
	;addc	a, #'0'
	;lcall	cout
	;clr	a
	;mov	c, pb_bit_4
	;addc	a, #'0'
	;lcall	cout
	;clr	a
	;mov	c, pb_bit_5
	;addc	a, #'0'
	;lcall	cout
	;clr	a
	;mov	c, pb_bit_6
	;addc	a, #'0'
	;lcall	cout
	;clr	a
	;mov	c, pb_bit_7
	;addc	a, #'0'
	;lcall	cout
	;clr	a
	;mov	c, pb_bit_8
	;addc	a, #'0'
	;lcall	cout
	;clr	a
	;mov	c, pb_bit_9
	;addc	a, #'0'
	;lcall	cout
	;clr	a
	;mov	c, pb_bit_A
	;addc	a, #'0'
	;lcall	cout
	;clr	a
	;mov	c, pb_bit_B
	;addc	a, #'0'
	;lcall	cout
	;lcall	newline

	pop	acc
	pop	psw
	reti



	


;***********************************************************************
;**                                                                   **
;**                       Serial Port Routines                        **
;**                                                                   **
;***********************************************************************


init_uart_intr:
	clr	ea
        orl     pcon, #10000000b	;set double baud rate
	anl	tmod, #00001111b	;clear all timer1 bits in tmod
	orl	tmod, #00100000b	;set timer1 as 8 bit auto reload
	clr	tr1			;make sure timer1 isn't running
	clr	tf1
        mov     a, #baud_const
        mov     th1, a			;set timer1 rate
        mov     tl1, a
	mov	scon, #01010000b	;config serial port (ri and ti cleared)
        mov     rx_buf_head, #rx_buf	;set up empty receive buffer
        mov     rx_buf_tail, #rx_buf
        mov     tx_buf_head, #0		;set up transmit buffer as empty
        mov     tx_buf_tail, #0		;and no tx interrupt expected
	setb	tr1			;start timer1
        setb    es			;enable serial port interrupt
        setb    ea			;also turn on interrupts in general
        ret


;service interrupts generated by the uart... put a received character
;into the receive buffer or get a character from the transmit buffer
;and send it.

uart_intr:
        push    psw
        push    acc
        jb      ti, xmit
recv:   clr     ri
        mov     a, rx_buf_head
        inc     a
        cjne    a, #rx_buf+rx_buf_size, recv2   ;check for wrap around
        mov     a, #rx_buf
recv2:  cjne    a, rx_buf_tail, recv_ok         ;check for buffer full
        ;If we got here, the recv buffer is full, so we just
        ;discard this character.  Maybe there should be an overrun
        ;flag so the main program can know some data was lost...
        mov     a, sbuf
        pop     acc
        pop     psw
        reti
recv_ok:mov     rx_buf_head, a  ;update rx_buf_head while it's still in Acc.
        xch     a, r0           ;put r0 onto the stack, since we need @r0 ptr
        push    acc
        mov     a, sbuf         ;get the incoming data
        mov     @r0, a          ;and put it in the buffer
        pop     acc
        mov     r0, a
        pop     acc
        pop     psw
        reti
xmit:   clr     ti
        mov     a, tx_buf_tail
        cjne    a, tx_buf_head, xmit2           ;check for buffer empty
        ;If we got here, there is no data waiting to transmit.  The
        ;uart won't generate more interrupts, so the cout routine had
        ;better check set ti.  tx_buf_head is loaded with zero to that
        ;cout will know to set ti.
        mov     tx_buf_head, #0
        pop     acc
        pop     psw
        reti
xmit2:  inc     a
        cjne    a, #tx_buf+tx_buf_size, xmit3   ;check for wrap around
        mov     a, #tx_buf
xmit3:  mov     tx_buf_tail, a  ;update tx_buf_tail while it's still in Acc.
        xch     a, r0           ;put r0 into acc, since we use @r0 as a pointer
        mov     sbuf, @r0
        mov     r0, a
        pop     acc
        pop     psw
        reti


;Get a character from the receive buffer.  If nothing is in the
;buffer, wait for something to appear.

cin:	setb	cin_serviced
	clr     es
        mov     a, rx_buf_tail
        cjne    a, rx_buf_head, cin2    ;check if buffer empty
        setb    es
        nop
        sjmp    cin
cin2:   inc     a
        cjne    a, #rx_buf+rx_buf_size, cin3    ;check for wrap around
        mov     a, #rx_buf
cin3:   mov     rx_buf_tail, a
        xch     a, r0
        push    acc                     ;keep r0 value on the stack
        mov     a, @r0
        setb    es
        mov     r0, a
        pop     acc
        xch     a, r0
        ret

;Put the character in Acc into the transmit buffer.  If the buffer is
;full, wait until these is a space to put it.

cout:
        push    psw
        xch     a, r0                           ;put char to send into r0
        push    acc                             ;keep r0 on stack
cout_wt:clr     es
        mov     a, tx_buf_head
        jnz     cout2                           ;check if not transmitting
        setb    ti              ;set ti in software since buffer is empty
        mov     a, #tx_buf
        mov     tx_buf_head, a
        mov     tx_buf_tail, a
cout2:  inc     a
        cjne    a, #tx_buf+tx_buf_size, cout3
        mov     a, #tx_buf
cout3:  cjne    a, tx_buf_tail, cout4           ;check if buffer full
        setb    es
        nop                                     ;wait for space in buffer
        sjmp    cout_wt
cout4:  mov     tx_buf_head, a                  ;update tx_buf_head
        xch     a, r0           ;char to send in Acc, tx_buf_head in r0.
        mov     @r0, a
        setb    es
        xch     a, r0                           ;restore r0 value and Acc
        pop     acc
        xch     a, r0
        pop     psw
        ret

; Highly code efficient resursive call phex contributed
; by Alexander B. Alexandrov <abalex@cbr.spb.ru>

phex:
phex8:
        acall   phex_b
phex_b: swap    a               ;SWAP A will be twice => A unchanged
phex1:  push    acc
        anl     a, #15
        add     a, #0x90        ; acc is 0x9X, where X is hex digit
        da      a               ; if A to F, C=1 and lower four bits are 0..5
        addc    a, #0x40
        da      a
        acall   cout
        pop     acc
        ret


newline:push    acc             ;print one newline
        mov     a, #13
        lcall   cout
        mov     a, #10
        lcall   cout
        pop     acc
        ret

pstr:	clr	a
	movc	a, @a+dptr
	inc	dptr
	jz	pstr2
	lcall	cout
	sjmp	pstr
pstr2:	ret


;***********************************************************************
;**                                                                   **
;**                   Fonts and Font Manipulation                     **
;**                                                                   **
;***********************************************************************


	;use the character in Acc and "font" to lookup the 5 byte
	;data to draw a character, and copy it into "char_buf".
lookup_char:
	push	acc
	mov	b, #5
	mul	ab
	mov	dpl, a		;compute offset within font
	mov	dph, b
	mov	a, font
	mov	b, #5
	mul	ab		;compute font's location
	add	a, dph
	mov	dph, a		;dptr points to the character we need
	movx	a, @dptr
	mov	char_buf+0, a
	inc	dptr
	movx	a, @dptr
	mov	char_buf+1, a
	inc	dptr
	movx	a, @dptr
	mov	char_buf+2, a
	inc	dptr
	movx	a, @dptr
	mov	char_buf+3, a
	inc	dptr
	movx	a, @dptr
	mov	char_buf+4, a
	pop	acc
	ret



test_font:
	;r6 has the first character to draw
	push	mode_bits
	clr	scroll_mode
	clr	wrap_mode
	lcall	init_display
	mov	r4, #0
tfont2:
	mov	r5, #16
	mov	x_pos, #8
	mov	y_pos, r4
tfont3:
	mov	a, r6
	push	acc
	lcall	lookup_char
	lcall	draw_char
	pop	acc
	mov	r6, a
	inc	r6
	djnz	r5, tfont3
	inc	r4
	cjne	r4, #8, tfont2
	pop	mode_bits
	ret



copy_fonts_to_sram:
	mov	dptr, #fonts	;dptr = source address
	mov	r0, #0		;r0/r1 = destination address
	mov	r1, #0
cpy_fonts_loop1:
	mov	p2, r1
cpy_fonts_loop2:
	clr	a
	movc	a, @a+dptr
	inc	dptr
	movx	@r0, a
	inc	r0
	cjne	r0, #0, cpy_fonts_loop2
	inc	r1
	cjne	r1, #20, cpy_fonts_loop1
	mov	p2, #255
	ret


msg_font:
	.db	"Font #",0
msg_f0:
	.db	"0 to",0
	.db	"127",0
msg_f128:
	.db	"128 to",0
	.db	"255",0


mesg_abort:
	.db	"Press ESC twice to return to PAULMON2",13,10,13,10,0


fonts:
	; this must be the last label of the file, because the build
	; process in the makefile will add the fonts after this point

